/*
 * True random number computation and storage
 *
 * (c) Copyright 1990-1994 by Philip Zimmermann.  All rights reserved.
 * The author assumes no liability for damages resulting from the use
 * of this software, even if the damage results from defects in this
 * software.  No warranty is expressed or implied.
 *
 * Note that while most PGP source modules bear Philip Zimmermann's
 * copyright notice, many of them have been revised or entirely written
 * by contributors who frequently failed to put their names in their
 * code.  Code that has been incorporated into PGP from other authors
 * was either originally published in the public domain or is used with
 * permission from the various authors.
 *
 * PGP is available for free to the public under certain restrictions.
 * See the PGP User's Guide (included in the release package) for
 * important information about licensing, patent restrictions on
 * certain algorithms, trademarks, copyrights, and export controls.
 *
 * Written by Colin Plumb.
 *
 * 2.5n  May 10, 1994  G. K. Pace
 * Changed MSDOS to __MSDOS__, added __OS2__.
 */
#include <stdlib.h>

/* gkp */
#include <mem.h>


#include "randpool.h"
#include "usuals.h"
#include "md5.h"

/* The pool must be a multiple of the 16-byte (128-bit) MD5 block size */
#define RANDPOOLBYTES ((RANDPOOLBITS+127 & ~127) >> 3)
#if RANDPOOLBYTES <= 64
#error Random pool too small - please increase RANDPOOLBITS in randpool.h
#endif

static byte randPool[RANDPOOLBYTES];	/* Random pool */
static unsigned randPoolGetPos = 0;	/* Current position to get from */
static unsigned randPoolAddPos = 0;	/* Current position to add to */

static byte randPoolKey[64];

static void
xorbytes(byte *dest, byte const *src, unsigned len)
{
	while (len--)
		*dest++ ^= *src++;
}

/*
 * Destroys already-used random numbers.  Ensures no sensitive data
 * remains in memory that can be recovered later.  This is also
 * called to "stir in" newly acquired environmental noise bits.
 * These noise bits are placed in the trueRandKey.
 *
 * The transformation is carried out by "encrypting" the data in CBC
 * mode with MD5 as the block cipher.  This is not invertible, but that
 * is just fine for these purposes.
 *
 * Then, to make doubly certain the stirring operation is strictly one-way,
 * we destroy the key.  This is done by reinitializing it from the pool,
 * copying the first 64 bytes of the pool over the key.  These bytes are
 * not returned by randPoolGetBytes().
 */
void
randPoolStir(void)
{
	int i;

	xorbytes(randPool, randPool+sizeof(randPool)-16, 16);
	Transform((word32 *)randPool, (word32 *)randPoolKey);
	for(i = 16; i < sizeof(randPool); i += 16) {
		xorbytes(randPool+i, randPool+i-16, 16);
		Transform((word32 *)(randPool+i), (word32 *)randPoolKey);
	}

	memcpy(randPoolKey, randPool, sizeof(randPoolKey));

	randPoolAddPos = 0;
	randPoolGetPos = sizeof(randPoolKey);
}

void
randPoolAddBytes(byte const *buf, unsigned len)
{
	unsigned t;

	while (len > (t = sizeof(randPool) - randPoolAddPos)) {
		xorbytes(randPool+randPoolAddPos, buf, t);
		buf += t;
		len -= t;
		randPoolStir();
	}

	if (len) {
		xorbytes(randPool+randPoolAddPos, buf, len);
		randPoolAddPos += len;
		randPoolGetPos = sizeof(randPool);
	}
}

void
randPoolGetBytes(byte *buf, unsigned len)
{
	unsigned t;

	while (len > (t = sizeof(randPool) - randPoolGetPos)) {
		memcpy(buf, randPool+randPoolGetPos, t);
		buf += t;
		len -= t;
		randPoolStir();
	}

	if (len) {
		memcpy(buf, randPool+randPoolGetPos, len);
		randPoolGetPos += len;
	}
}

byte
randPoolGetByte(void)
{
	if (randPoolGetPos == sizeof(randPool))
		randPoolStir();

	return randPool[randPoolGetPos++];
}

